#%% [markdown]
# # One-dimensional Examples
#
# This jupyter notebook contains two toy examples for integration problems that benefit from our approach: stratified sampling and defensive sampling.

#%%
import numpy as np
import matplotlib.pyplot as plt
import scipy.integrate as integrate

gauss = lambda x, mu, sigma: np.exp(-(x-mu)**2/(2 * sigma**2))

#%% [markdown]
#  ## Case 1: Stratified sampling
#  In this example, we show how our approach can account for variance reduction
#  (or increase) due to methods that do not alter the effective density.
#  We illustrate this on a toy example, using an unstratified and a stratified uniform
#  distribution for MC integration of a simple exponential function.
#%% [markdown]
# ### Set-up
#  The integrand is a simple exponential function.

#%%
domain = [0, 2]
f = lambda x: gauss(x, 0, 0.5) + 0.25
expectedValue = integrate.quad(f, domain[0], domain[1])[0]

#%% [markdown]
#  The first technique is stratified, it uniformly draws one sample within
#  equal-sized intervals of the domain.
#  The second technique uniformly samples the entire domain, but using
#  a number of samples equal to the number of strata from the other technique.

#%%
numStrata = 4
stratSize = (domain[1] - domain[0]) / numStrata
pdfStrat = lambda x: 1 / stratSize * np.ones(np.array(x).shape)
pdfUnstrat = lambda x: 1 / (domain[1] - domain[0]) * np.ones(np.array(x).shape)

plt.figure()
x = np.linspace(domain[0], domain[1], 1000)
plt.plot(x, f(x), label='$f(x)$')
plt.plot(x, pdfStrat(x), label='$p_{strat}(x)$')
plt.plot(x, pdfUnstrat(x), label='$p_{unstrat}(x)$')
plt.legend()

plt.vlines(np.arange(numStrata+1) * stratSize, ymin=0, ymax=2, linestyles='dashed')

plt.savefig('stratproblem.pdf')

#%% [markdown]
# ### Variance and weight comparisons
#  Since the strata are estimated independently, the variance of their sum
#  is the sum of the individual variances.

#%%
varStrat = 0
for s in range(numStrata):
    lo = domain[0] + s * stratSize
    hi = lo + stratSize
    secondMoment = integrate.quad(lambda x: f(x)**2 / pdfStrat(x), lo, hi)[0]
    firstMomentSquare = integrate.quad(lambda x: f(x), lo, hi)[0]**2
    varStrat += secondMoment - firstMomentSquare

print('Stratified variance: ' + str(varStrat))

#%% [markdown]
#  The unstratified estimator uses numStrata independent samples, therfore its variance is

#%%
varUnstrat = 0
secondMoment = integrate.quad(lambda x: f(x)**2 / pdfUnstrat(x), domain[0], domain[1])[0]
firstMomentSquare = integrate.quad(lambda x: f(x), domain[0], domain[1])[0]**2
varUnstrat = (secondMoment - firstMomentSquare) / numStrata

print('Unstratified variance: ' + str(varUnstrat))

#%% [markdown]
#  Now we compute the factors needed for our method.

#%%
factorStrat = 1 + expectedValue**2 / varStrat
factorUnstrat = 1 + expectedValue**2 / varUnstrat
print(factorStrat)
print(factorUnstrat)

#%% [markdown]
#  With equal effective densities, our method also yields constant weights
#  (when applied globally, as done here). For this use-case, we hence additionally compare
#  to the optimal constant weight (proportional to the reciporcal variances).

#%%
wStratVeach = 0.5
wUnstratVeach = 0.5

wStratOur = factorStrat / (factorStrat + factorUnstrat)
wUnstratOur = factorUnstrat / (factorStrat + factorUnstrat)

wStratVar = varStrat**(-1) / (varStrat**(-1) + varUnstrat**(-1))
wUnstratVar = varUnstrat**(-1) / (varStrat**(-1) + varUnstrat**(-1))

fig, ax = plt.subplots(nrows=1, ncols=3)
ax[0].bar(np.arange(2), [wStratVeach, wUnstratVeach])
ax[0].set_ylim([0,1])
ax[1].bar(np.arange(2), [wStratOur, wUnstratOur])
ax[1].set_ylim([0,1])
ax[2].bar(np.arange(2), [wStratVar, wUnstratVar])
ax[2].set_ylim([0,1])

#%% [markdown]
#  Computation of the optimal weights (Kondapaneni et al 2019) is particularly
#  tricky in this example.
#  Their derivation uses the assumption that the pdf integrates to one over the
#  whole integration domain.
#  When considering stratified techniques, this means that we cannot treat the
#  problem as a combination of two techniques.
#  In fact, the stratified "technique" consists of $N$ techniques, each sampling
#  exclusively within one stratum.
#
#  (Trying to still compute the values from Kondapaneni et al 2019, pretending
#  we were combining just two techniques, results in the same variance as the
#  balance heuristic.)

#%%
techniqueMatrix = np.zeros((numStrata+1,numStrata+1))
techniqueMatrix[0,0] = 1/(2 * numStrata)
for i in range(1,numStrata+1):
    techniqueMatrix[i,0] = stratSize / numStrata * 0.5
    techniqueMatrix[0,i] = stratSize / numStrata * 0.5
    techniqueMatrix[i,i] = 0.5

contribVector = np.zeros(numStrata+1)
contribVector[0] = 1/(2 * numStrata) * expectedValue
for i in range(1, numStrata+1):
    lo = domain[0] + (i-1) * stratSize
    hi = lo + stratSize
    mean = integrate.quad(lambda x: f(x), lo, hi)[0]
    contribVector[i] = 0.5 * mean

optimalIntegrals = np.linalg.solve(techniqueMatrix, contribVector)
densities = [pdfUnstrat]
densities += [pdfStrat for i in range(numStrata)]


#%%
def weightOptimalUnstrat(x):
    # find correspondig strata for all x
    idx = np.floor((x-domain[0])/stratSize).astype(int)
    idx = np.clip(idx, 0, numStrata-1) + 1

    xs = np.array(x)
    weightedDensities = np.zeros(xs.shape)
    weightedDensities += optimalIntegrals[0] * densities[0](x)
    for i in range(1,numStrata+1):
        mask = idx == i
        weightedDensities[mask] += optimalIntegrals[i] * densities[i](xs[mask])

    return optimalIntegrals[0] * densities[0](x) / f(x) + 0.5 * (1 - weightedDensities / f(x))

def weightOptimalStrat(x):
    # find correspondig strata for all x
    idx = np.floor((x-domain[0])/stratSize).astype(int)
    idx = np.clip(idx, 0, numStrata-1) + 1

    xs = np.array(x)
    weightedDensities = np.zeros(xs.shape)
    weightedDensities += optimalIntegrals[0] * densities[0](x)

    for i in range(1,numStrata+1):
        mask = idx == i
        weightedDensities[mask] += optimalIntegrals[i] * densities[i](xs[mask])
    w = np.zeros(np.array(x).shape)
    if xs.shape == ():
        return optimalIntegrals[idx] * densities[idx](x) / f(x) + 0.5 * (1 - weightedDensities / f(x))

    for k in range(xs.shape[0]):
        w[k] = optimalIntegrals[idx[k]] * densities[idx[k]](x[k]) / f(x[k]) + 0.5 * (1 - weightedDensities[k] / f(x[k]))
    return w



#%% [markdown]
#  The variance using the optimal weights would be:

#%%
varStratOpt = 0
for s in range(numStrata):
    lo = domain[0] + s * stratSize
    hi = lo + stratSize
    secondMoment = integrate.quad(lambda x: weightOptimalStrat(x)**2 * f(x)**2 / pdfStrat(x), lo, hi)[0]
    firstMomentSquare = integrate.quad(lambda x: weightOptimalStrat(x) * f(x), lo, hi)[0]**2
    varStratOpt += secondMoment - firstMomentSquare

varUnstratOpt = 0
secondMoment = integrate.quad(lambda x: weightOptimalUnstrat(x)**2 * f(x)**2 / pdfUnstrat(x), domain[0], domain[1])[0]
firstMomentSquare = integrate.quad(lambda x: weightOptimalUnstrat(x) * f(x), domain[0], domain[1])[0]**2
varUnstratOpt = (secondMoment - firstMomentSquare) / numStrata

varOptimal = varStratOpt + varUnstratOpt

#%% [markdown]
#  Finally, we compute and compare the variances of the other approaches.
#  This is greatly simplified by the fact that all weighting functions are constant.

#%%
varVeach = (wStratVeach**2 * varStrat + wUnstratVeach**2 * varUnstrat)
varOur = (wStratOur**2 * varStrat + wUnstratOur**2 * varUnstrat)
varVar = (wStratVar**2 * varStrat + wUnstratVar**2 * varUnstrat)


print('Balance/power variance: ' + str(varVeach) + ' (' + str(varVeach / varStrat) + ' times that of the stratified alone)')
print('Our variance: ' + str(varOur) + ' (' + str(varOur / varStrat) + ' times that of the stratified alone)')
print('Optimal (constant) variance: ' + str(varVar) + ' (' + str(varVar / varStrat) + ' times that of the stratified alone)')
print('Optimal (non-constant) variance: ' + str(varOptimal) + ' (' + str(varOptimal / varStrat) + ' times that of the stratified alone)')


fig, ax = plt.subplots(nrows=1, ncols=7, figsize=(24,3))

x = np.linspace(domain[0], domain[1], 1000)
ax[0].plot(x, pdfStrat(x), label='$p_{strat}(x)$')
ax[0].plot(x, pdfUnstrat(x), label='$p_{unstrat}(x)$')
ax[0].plot(x, f(x), label='$f(x)$')
ax[0].legend()
ax[0].set_yticks([0,1,2])

ax[0].vlines(np.arange(numStrata+1) * stratSize, ymin=0, ymax=2, linestyles='dashed')

ax[1].plot(x, wStratVeach * np.ones(x.shape), label='strat')
ax[1].plot(x, wUnstratVeach * np.ones(x.shape), label='unstrat')
ax[1].set_ylim([0,1])
ax[1].set_yticks([0,0.5,1])

ax[2].plot(x, wStratVeach * np.ones(x.shape), label='strat')
ax[2].plot(x, wUnstratVeach * np.ones(x.shape), label='unstrat')
ax[2].set_ylim([0,1])
ax[2].set_yticks([0,0.5,1])

ax[3].plot(x, wStratVar * np.ones(x.shape), label='strat')
ax[3].plot(x, wUnstratVar * np.ones(x.shape), label='unstrat')
ax[3].set_ylim([0,1])
ax[3].set_yticks([0,0.5,1])

ax[4].plot(x, wStratOur * np.ones(x.shape), label='strat')
ax[4].plot(x, wUnstratOur * np.ones(x.shape), label='unstrat')
ax[4].set_ylim([0,1])
ax[4].set_yticks([0,0.5,1])

ax[5].plot(x, weightOptimalStrat(x), label='strat')
ax[5].plot(x, weightOptimalUnstrat(x), label='unstrat')
ax[5].set_ylim([-0.2,1.2])
ax[5].set_yticks([0,0.5,1])
ax[5].legend()

ax[6].bar(np.arange(5),
    np.array([varVeach, varVeach, varVar, varOur, varOptimal]))
ax[6].set_yticks(np.array([varVeach, varOur, varOptimal]))
ax[6].set_xticks(np.arange(5))
ax[6].set_xticklabels(['balance', 'power', 'variance', 'our', 'optimal'],
    rotation=45, ha='right')

plt.savefig('onedimstrat.pdf')

#%% [markdown]
#  ## Case 2: Defensive sampling
#  Previous work has shown that the balance heuristic can perform poorly
#  in defensive sampling cases, as these are likely to trigger the problematic
#  "low-variance" scenario.
#  The power heuristic, the favored solution for these cases, does not always perform
#  better. It can overshoot and increase the variance even further, as the following
#  example shows.
#%% [markdown]
#  ### Set-up
#  The integrand is a simple exponential function.

#%%
domain = [0, 2]
f = lambda x: gauss(x, 0, 0.5) + 0.25
expectedValue = integrate.quad(f, domain[0], domain[1])[0]

#%% [markdown]
#  To model a defensive sampling use-case, we use as the first density a slightly
#  shifted and scaled version of the integrand.

#%%
targetFn = lambda x: gauss(x, 0.01, 0.65) + 0.125
mass = integrate.quad(targetFn, domain[0], domain[1])[0]
pdfLearned = lambda x: targetFn(x) / mass

#%% [markdown]
#  The second density is the defensive one, here a uniform one.

#%%
pdfDefensive = lambda x: 1/(domain[1] - domain[0]) * np.ones(np.array(x).shape)

#%% [markdown]
#  The following plot visualizes the scenario

#%%
plt.figure()
x = np.linspace(domain[0], domain[1], 1000)
plt.plot(x, pdfLearned(x), label='$p_{learned}(x)$')
plt.plot(x, pdfDefensive(x), label='$p_{defensive}(x)$')
plt.plot(x, f(x), label='$f(x)$')
plt.legend()

plt.savefig('defproblem.pdf')

#%% [markdown]
#  ### Combination strategies
#  Now we plot the MIS weights with
#  the balance and power heuristic,
#  the optimal weights (Kondapaneni et al 2019),
#  and our approach.

#%%
weightBalanceLearned = lambda x: pdfLearned(x) / (pdfDefensive(x) + pdfLearned(x))
weightBalanceDefensive = lambda x: pdfDefensive(x) / (pdfDefensive(x) + pdfLearned(x))

weightPowerLearned = lambda x: pdfLearned(x)**2 / (pdfDefensive(x)**2 + pdfLearned(x)**2)
weightPowerDefensive = lambda x: pdfDefensive(x)**2 / (pdfDefensive(x)**2 + pdfLearned(x)**2)

fig, ax = plt.subplots(nrows=1, ncols=2)
ax[0].plot(x, weightBalanceLearned(x))
ax[0].plot(x, weightBalanceDefensive(x))

ax[1].plot(x, weightPowerLearned(x))
ax[1].plot(x, weightPowerDefensive(x))

ax[0].set_ylim([0,1])
ax[1].set_ylim([0,1])

#%% [markdown]
#  Now we compute the optimal weights (Kondapaneni et al 2019).
#  First, we construct and solve their linear system.

#%%
techniqueMatrix = np.zeros((2, 2))
contribVector = np.zeros(2)

weightBalance = [weightBalanceLearned, weightBalanceDefensive]
densities = [pdfLearned, pdfDefensive]

for i in range(2):
    contribVector[i] = integrate.quad(lambda x:
        weightBalance[i](x) * f(x),
        domain[0], domain[1])[0]
    for j in range(2):
        techniqueMatrix[i,j] = integrate.quad(lambda x:
            densities[i](x) * weightBalance[j](x),
            domain[0], domain[1])[0]

optimalIntegrals = np.linalg.solve(techniqueMatrix, contribVector)

#%% [markdown]
#  Now, we compute the actual optimal weights based on the optimally weighted integrals.

#%%
def weightOptimalLearned(x):
    weightedDensities = 0.0
    for i in range(2):
        weightedDensities += optimalIntegrals[i] * densities[i](x)
    return optimalIntegrals[0] * densities[0](x) / f(x) + weightBalance[0](x) * (1 - (weightedDensities) / f(x))

def weightOptimalDefensive(x):
    weightedDensities = 0.0
    for i in range(2):
        weightedDensities += optimalIntegrals[i] * densities[i](x)
    return optimalIntegrals[1] * densities[1](x) / f(x) + weightBalance[1](x) * (1 - (weightedDensities) / f(x))

fig, ax = plt.subplots(nrows=1, ncols=1)
ax.plot(x, weightOptimalLearned(x))
ax.plot(x, weightOptimalDefensive(x))
ax.set_ylim([0,1])


#%%
varianceBalance = 0

secondMoment = integrate.quad(lambda x: weightBalanceLearned(x)**2 * f(x)**2 / pdfLearned(x),
                              domain[0], domain[1])[0]
firstMomentSquare = integrate.quad(lambda x: weightBalanceLearned(x) * f(x),
                                   domain[0], domain[1])[0]**2
varianceBalance += secondMoment - firstMomentSquare

secondMoment = integrate.quad(lambda x: weightBalanceDefensive(x)**2 * f(x)**2 / pdfDefensive(x),
                              domain[0], domain[1])[0]
firstMomentSquare = integrate.quad(lambda x: weightBalanceDefensive(x) * f(x),
                                   domain[0], domain[1])[0]**2
varianceBalance += secondMoment - firstMomentSquare

print('Balance heuristic: ' + str(varianceBalance) + ' (' + str(varianceBalance / varianceBalance) + ' times balance)')


#%%

variancePower = 0

secondMoment = integrate.quad(lambda x: weightPowerLearned(x)**2 * f(x)**2 / pdfLearned(x),
                              domain[0], domain[1])[0]
firstMomentSquare = integrate.quad(lambda x: weightPowerLearned(x) * f(x),
                                   domain[0], domain[1])[0]**2
variancePower += secondMoment - firstMomentSquare

secondMoment = integrate.quad(lambda x: weightPowerDefensive(x)**2 * f(x)**2 / pdfDefensive(x),
                              domain[0], domain[1])[0]
firstMomentSquare = integrate.quad(lambda x: weightPowerDefensive(x) * f(x),
                                   domain[0], domain[1])[0]**2
variancePower += secondMoment - firstMomentSquare

print('Power heuristic: ' + str(variancePower) + ' (' + str(variancePower / varianceBalance) + ' times balance)')


#%%
varianceOptimal = 0

secondMoment = integrate.quad(lambda x: weightOptimalLearned(x)**2 * f(x)**2 / pdfLearned(x),
                              domain[0], domain[1])[0]
firstMomentSquare = integrate.quad(lambda x: weightOptimalLearned(x) * f(x),
                                   domain[0], domain[1])[0]**2
varianceOptimal += secondMoment - firstMomentSquare

secondMoment = integrate.quad(lambda x: weightOptimalDefensive(x)**2 * f(x)**2 / pdfDefensive(x),
                              domain[0], domain[1])[0]
firstMomentSquare = integrate.quad(lambda x: weightOptimalDefensive(x) * f(x),
                                   domain[0], domain[1])[0]**2
varianceOptimal += secondMoment - firstMomentSquare

print('Optimal: ' + str(varianceOptimal) + ' (' + str(varianceOptimal / varianceBalance) + ' times balance)')

#%% [markdown]
#  Compute the second moments and variances needed for our approach

#%%
firstMomentSquare = integrate.quad(lambda x: f(x), domain[0], domain[1])[0]**2

secondMomentLearned = integrate.quad(lambda x: f(x)**2 / pdfLearned(x), domain[0], domain[1])[0]
ourFactorLearned = secondMomentLearned / (secondMomentLearned - firstMomentSquare)

secondMomentDefensive = integrate.quad(lambda x: f(x)**2 / pdfDefensive(x), domain[0], domain[1])[0]
ourFactorDefensive = secondMomentDefensive / (secondMomentDefensive - firstMomentSquare)

weightOurLearned = lambda x: pdfLearned(x) * ourFactorLearned / (pdfDefensive(x) * ourFactorDefensive + pdfLearned(x) * ourFactorLearned)
weightOurDefensive = lambda x: pdfDefensive(x) * ourFactorDefensive / (pdfDefensive(x) * ourFactorDefensive + pdfLearned(x) * ourFactorLearned)

fig, ax = plt.subplots(nrows=1, ncols=1)
ax.plot(x, weightOurLearned(x))
ax.plot(x, weightOurDefensive(x))
ax.set_ylim([0,1])

#%% [markdown]
#  Compute the variance when combining using our method

#%%
varianceOur = 0

secondMoment = integrate.quad(lambda x: weightOurLearned(x)**2 * f(x)**2 / pdfLearned(x),
                              domain[0], domain[1])[0]
firstMomentSquare = integrate.quad(lambda x: weightOurLearned(x) * f(x),
                                   domain[0], domain[1])[0]**2
varianceOur += secondMoment - firstMomentSquare

secondMoment = integrate.quad(lambda x: weightOurDefensive(x)**2 * f(x)**2 / pdfDefensive(x),
                              domain[0], domain[1])[0]
firstMomentSquare = integrate.quad(lambda x: weightOurDefensive(x) * f(x),
                                   domain[0], domain[1])[0]**2
varianceOur += secondMoment - firstMomentSquare

print('Our heuristic: ' + str(varianceOur) + ' (' + str(varianceOur / varianceBalance) + ' times balance)')

#%% [markdown]
# Compute the optimal constant heuristic results. The optimal constant weights are proprotional to the reciprocal variances of the estimators.

#%%
recipVarLearned = 1 / (secondMomentLearned - firstMomentSquare)
recipVarDefensive = 1 / (secondMomentDefensive - firstMomentSquare)

weightOptConstLearned = lambda x: recipVarLearned / (recipVarDefensive + recipVarLearned) * np.ones(np.array(x).shape)
weightOptConstDefensive = lambda x: recipVarDefensive / (recipVarDefensive + recipVarLearned) * np.ones(np.array(x).shape)

fig, ax = plt.subplots(nrows=1, ncols=1)
ax.plot(x, weightOptConstLearned(x))
ax.plot(x, weightOptConstDefensive(x))
ax.set_ylim([0,1])

#%% [markdown]
# Variance when using the optimal constant weight.

#%%
varianceOptimalConst = 0

secondMoment = integrate.quad(lambda x: weightOptConstLearned(x)**2 * f(x)**2 / pdfLearned(x),
                              domain[0], domain[1])[0]
firstMomentSquare = integrate.quad(lambda x: weightOptConstLearned(x) * f(x),
                                   domain[0], domain[1])[0]**2
varianceOur += secondMoment - firstMomentSquare

secondMoment = integrate.quad(lambda x: weightOptConstDefensive(x)**2 * f(x)**2 / pdfDefensive(x),
                              domain[0], domain[1])[0]
firstMomentSquare = integrate.quad(lambda x: weightOptConstDefensive(x) * f(x),
                                   domain[0], domain[1])[0]**2
varianceOptimalConst += secondMoment - firstMomentSquare

print('Optimal constant: ' + str(varianceOptimalConst) + ' (' + str(varianceOptimalConst / varianceBalance) + ' times balance)')


#%%
plt.figure()
plt.bar(np.arange(5), [varianceBalance,
                       variancePower,
                       varianceOur,
                       varianceOptimal,
                       varianceOptimalConst])
plt.xticks(np.arange(5), ['balance', 'power', 'our', 'optimal', 'optimal constant'])

plt.savefig('defvars.pdf')


#%%
fig, ax = plt.subplots(nrows=1, ncols=5, figsize=(15,3))
ax[0].plot(x, weightBalanceLearned(x))
ax[0].plot(x, weightBalanceDefensive(x))

ax[1].plot(x, weightPowerLearned(x))
ax[1].plot(x, weightPowerDefensive(x))

ax[2].plot(x, weightOptConstLearned(x))
ax[2].plot(x, weightOptConstDefensive(x))
ax[2].set_ylim([0,1])

ax[3].plot(x, weightOptimalLearned(x))
ax[3].plot(x, weightOptimalDefensive(x))
ax[3].set_ylim([0,1])

ax[4].plot(x, weightOurLearned(x))
ax[4].plot(x, weightOurDefensive(x))
ax[4].set_ylim([0,1])

ax[0].set_ylim([0,1])
ax[1].set_ylim([0,1])

plt.savefig('defweights.pdf')



#%%

fig, ax = plt.subplots(nrows=1, ncols=7, figsize=(24,3))

x = np.linspace(domain[0], domain[1], 1000)
ax[0].plot(x, pdfLearned(x), label='$p_{learned}(x)$')
ax[0].plot(x, pdfDefensive(x), label='$p_{defensive}(x)$')
ax[0].plot(x, f(x), label='$f(x)$')
ax[0].legend()
ax[0].set_yticks([0,0.6,1.2])

ax[1].plot(x, weightBalanceLearned(x))
ax[1].plot(x, weightBalanceDefensive(x))
ax[1].set_yticks([0,0.5,1])

ax[2].plot(x, weightPowerLearned(x))
ax[2].plot(x, weightPowerDefensive(x))
ax[2].set_yticks([0,0.5,1])

ax[3].plot(x, weightOptConstLearned(x))
ax[3].plot(x, weightOptConstDefensive(x))
ax[3].set_ylim([0,1])
ax[3].set_yticks([0,0.5,1])

ax[5].plot(x, weightOptimalLearned(x))
ax[5].plot(x, weightOptimalDefensive(x))
ax[5].set_ylim([0,1])
ax[5].set_yticks([0,0.5,1])

ax[4].plot(x, weightOurLearned(x))
ax[4].plot(x, weightOurDefensive(x))
ax[4].set_ylim([0,1])
ax[4].set_yticks([0,0.5,1])

ax[1].set_ylim([0,1])
ax[2].set_ylim([0,1])

ax[0].set_xticks([domain[0], domain[1]/2, domain[1]])
ax[1].set_xticks([domain[0], domain[1]/2, domain[1]])
ax[2].set_xticks([domain[0], domain[1]/2, domain[1]])
ax[3].set_xticks([domain[0], domain[1]/2, domain[1]])
ax[4].set_xticks([domain[0], domain[1]/2, domain[1]])
ax[5].set_xticks([domain[0], domain[1]/2, domain[1]])

ax[6].bar(np.arange(5), [varianceBalance,
                       variancePower,
                       varianceOptimalConst,
                       varianceOur,
                       varianceOptimal])
ax[6].set_xticks(np.arange(5))
ax[6].set_xticklabels(['balance', 'power', 'variance', 'our', 'optimal'],
    rotation=45, ha='right')
ax[6].set_yticks([varianceBalance,
                       variancePower,
                       varianceOptimalConst,
                       varianceOur,
                       varianceOptimal])

plt.savefig('onedimdef.pdf')



#%% [markdown]
# ## Case 3: Biased technique
# In this example, we combine a biased technique, sampling only a subset of the domain,
# with a defensive technique.

#%% [markdown]
#  ### Set-up
#  The integrand is a simple exponential function.

#%%
domain = [0, 2]
f = lambda x: gauss(x, 0, 0.5) + 0.25
expectedValue = integrate.quad(f, domain[0], domain[1])[0]

#%% [markdown]
#  The fist density is uniform, but non-zero only within two standard deviations.

#%%
pdfBiased = lambda x: (x < 0.5) * 2

#%% [markdown]
#  The second density is the defensive one, here a uniform one.

#%%
pdfDefensive = lambda x: 1/(domain[1] - domain[0]) * np.ones(np.array(x).shape)

#%% [markdown]
#  The following plot visualizes the scenario

#%%
plt.figure()
x = np.linspace(domain[0], domain[1], 1000)
plt.plot(x, pdfBiased(x), label='$p_{biased}(x)$')
plt.plot(x, pdfDefensive(x), label='$p_{defensive}(x)$')
plt.plot(x, f(x), label='$f(x)$')
plt.legend()


#%% [markdown]
#  ### Combination strategies
#  Now we plot the MIS weights with
#  the balance and power heuristic,
#  the optimal weights (Kondapaneni et al 2019),
#  and our approach.

#%%
weightBalanceBiased = lambda x: pdfBiased(x) / (pdfDefensive(x) + pdfBiased(x))
weightBalanceDefensive = lambda x: pdfDefensive(x) / (pdfDefensive(x) + pdfBiased(x))

weightPowerBiased = lambda x: pdfBiased(x)**2 / (pdfDefensive(x)**2 + pdfBiased(x)**2)
weightPowerDefensive = lambda x: pdfDefensive(x)**2 / (pdfDefensive(x)**2 + pdfBiased(x)**2)

fig, ax = plt.subplots(nrows=1, ncols=2)
ax[0].plot(x, weightBalanceBiased(x))
ax[0].plot(x, weightBalanceDefensive(x))

ax[1].plot(x, weightPowerBiased(x))
ax[1].plot(x, weightPowerDefensive(x))

ax[0].set_ylim([0,1])
ax[1].set_ylim([0,1])

#%% [markdown]
#  Now we compute the optimal weights (Kondapaneni et al 2019).
#  First, we construct and solve their linear system.

#%%
techniqueMatrix = np.zeros((2, 2))
contribVector = np.zeros(2)

weightBalance = [weightBalanceBiased, weightBalanceDefensive]
densities = [pdfBiased, pdfDefensive]

for i in range(2):
    contribVector[i] = integrate.quad(lambda x:
        weightBalance[i](x) * f(x),
        domain[0], domain[1])[0]
    for j in range(2):
        techniqueMatrix[i,j] = integrate.quad(lambda x:
            densities[i](x) * weightBalance[j](x),
            domain[0], domain[1])[0]

optimalIntegrals = np.linalg.solve(techniqueMatrix, contribVector)

#%% [markdown]
#  Now, we compute the actual optimal weights based on the optimally weighted integrals.

#%%
def weightOptimalBiased(x):
    weightedDensities = 0.0
    for i in range(2):
        weightedDensities += optimalIntegrals[i] * densities[i](x)
    return optimalIntegrals[0] * densities[0](x) / f(x) + weightBalance[0](x) * (1 - (weightedDensities) / f(x))

def weightOptimalDefensive(x):
    weightedDensities = 0.0
    for i in range(2):
        weightedDensities += optimalIntegrals[i] * densities[i](x)
    return optimalIntegrals[1] * densities[1](x) / f(x) + weightBalance[1](x) * (1 - (weightedDensities) / f(x))

fig, ax = plt.subplots(nrows=1, ncols=1)
ax.plot(x, weightOptimalBiased(x))
ax.plot(x, weightOptimalDefensive(x))
ax.set_ylim([0,1])


#%%
varianceBalance = 0

secondMoment = integrate.quad(lambda x: weightBalanceBiased(x)**2 * f(x)**2 / pdfBiased(x),
                              domain[0], 0.5)[0]
firstMomentSquare = integrate.quad(lambda x: weightBalanceBiased(x) * f(x),
                                   domain[0], 0.5)[0]**2
varianceBalance += secondMoment - firstMomentSquare

secondMoment = integrate.quad(lambda x: weightBalanceDefensive(x)**2 * f(x)**2 / pdfDefensive(x),
                              domain[0], domain[1])[0]
firstMomentSquare = integrate.quad(lambda x: weightBalanceDefensive(x) * f(x),
                                   domain[0], domain[1])[0]**2
varianceBalance += secondMoment - firstMomentSquare

print('Balance heuristic: ' + str(varianceBalance) + ' (' + str(varianceBalance / varianceBalance) + ' times balance)')


#%%

variancePower = 0

secondMoment = integrate.quad(lambda x: weightPowerBiased(x)**2 * f(x)**2 / pdfBiased(x),
                              domain[0], 0.5)[0]
firstMomentSquare = integrate.quad(lambda x: weightPowerBiased(x) * f(x),
                                   domain[0], 0.5)[0]**2
variancePower += secondMoment - firstMomentSquare

secondMoment = integrate.quad(lambda x: weightPowerDefensive(x)**2 * f(x)**2 / pdfDefensive(x),
                              domain[0], domain[1])[0]
firstMomentSquare = integrate.quad(lambda x: weightPowerDefensive(x) * f(x),
                                   domain[0], domain[1])[0]**2
variancePower += secondMoment - firstMomentSquare

print('Power heuristic: ' + str(variancePower) + ' (' + str(variancePower / varianceBalance) + ' times balance)')


#%%
varianceOptimal = 0

secondMoment = integrate.quad(lambda x: weightOptimalBiased(x)**2 * f(x)**2 / pdfBiased(x),
                              domain[0], 0.5)[0]
firstMomentSquare = integrate.quad(lambda x: weightOptimalBiased(x) * f(x),
                                   domain[0], 0.5)[0]**2
varianceOptimal += secondMoment - firstMomentSquare

secondMoment = integrate.quad(lambda x: weightOptimalDefensive(x)**2 * f(x)**2 / pdfDefensive(x),
                              domain[0], domain[1])[0]
firstMomentSquare = integrate.quad(lambda x: weightOptimalDefensive(x) * f(x),
                                   domain[0], domain[1])[0]**2
varianceOptimal += secondMoment - firstMomentSquare

print('Optimal: ' + str(varianceOptimal) + ' (' + str(varianceOptimal / varianceBalance) + ' times balance)')

#%% [markdown]
#  Compute the second moments and variances needed for our approach

#%%
firstMomentSquare = integrate.quad(lambda x: f(x), domain[0], domain[1])[0]**2
firstMomentSquareBiased = integrate.quad(lambda x: f(x), domain[0], 0.5)[0]**2

secondMomentBiased = integrate.quad(lambda x: f(x)**2 / pdfBiased(x), domain[0], 0.5)[0]
ourFactorBiased = 1#secondMomentBiased / (secondMomentBiased - firstMomentSquareBiased)

secondMomentDefensive = integrate.quad(lambda x: f(x)**2 / pdfDefensive(x), domain[0], domain[1])[0]
ourFactorDefensive = secondMomentDefensive / (secondMomentDefensive - firstMomentSquare)

weightOurBiased = lambda x: pdfBiased(x) * ourFactorBiased / (pdfDefensive(x) * ourFactorDefensive + pdfBiased(x) * ourFactorBiased)
weightOurDefensive = lambda x: pdfDefensive(x) * ourFactorDefensive / (pdfDefensive(x) * ourFactorDefensive + pdfBiased(x) * ourFactorBiased)


fig, ax = plt.subplots(nrows=1, ncols=1)
ax.plot(x, weightOurBiased(x))
ax.plot(x, weightOurDefensive(x))
ax.set_ylim([0,1])

#%% [markdown]
#  Compute the variance when combining using our method

#%%
varianceOur = 0

secondMoment = integrate.quad(lambda x: weightOurBiased(x)**2 * f(x)**2 / pdfBiased(x),
                              domain[0], 0.5)[0]
firstMomentSquare = integrate.quad(lambda x: weightOurBiased(x) * f(x),
                                   domain[0], 0.5)[0]**2
varianceOur += secondMoment - firstMomentSquare

secondMoment = integrate.quad(lambda x: weightOurDefensive(x)**2 * f(x)**2 / pdfDefensive(x),
                              domain[0], domain[1])[0]
firstMomentSquare = integrate.quad(lambda x: weightOurDefensive(x) * f(x),
                                   domain[0], domain[1])[0]**2
varianceOur += secondMoment - firstMomentSquare

print('Our heuristic: ' + str(varianceOur) + ' (' + str(varianceOur / varianceBalance) + ' times balance)')


#%%
